from assimilator.core.database import Specification


class RedisSpecification(Specification):
    def apply(self, query: str) -> str:
        return super(RedisSpecification, self).apply(query)


class RedisKeySpecification(RedisSpecification):
    def __init__(self, key: str):
        self.filters = key

    def apply(self, query: str) -> str:
        return f"{query}{self.filters}"



from sqlalchemy.exc import IntegrityError

from assimilator.core.database.unit_of_work import UnitOfWork
from assimilator.core.database.exceptions import InvalidQueryError


class AlchemyUnitOfWork(UnitOfWork):
    def begin(self):
        self.repository.session.begin()

    def rollback(self):
        self.repository.session.rollback()

    def close(self):
        self.repository.session.close()

    def commit(self):
        try:
            self.repository.session.commit()
        except IntegrityError as exc:
            raise InvalidQueryError(exc)



from abc import ABC

from assimilator.core.database.unit_of_work import UnitOfWork
from assimilator.core.events.events_bus import EventBus


class OutboxRelay(ABC):
    def __init__(self, uow: UnitOfWork, event_bus: EventBus):
        self.uow = uow
        self.event_bus = event_bus

    def start(self):
        raise NotImplementedError("start() is not implemented")

    def acknowledge(self, events):
        raise NotImplementedError("acknowledge() is not implemented")



import json
from typing import Iterable, Type

import redis

from assimilator.core.database.repository import BaseRepository
from assimilator.core.database.specifications import Specification
from assimilator.redis.database.models import RedisModel
from assimilator.redis.database.specifications import FilterSpecification


class RedisLazyCommand:
    def __init__(self, command: callable, *args):
        self.command = command
        self.args = args

    def __call__(self):
        return self.command(*self.args)


class RedisRepository(BaseRepository):
    def __init__(
        self,
        session: redis.Redis,
        model: Type[RedisModel],
        lazy_command_cls: Type[RedisLazyCommand] = RedisLazyCommand,
    ):
        super(RedisRepository, self).__init__(session)
        self.lazy_command_cls = lazy_command_cls
        self.model = model

    def get_initial_query(self):
        """
        Redis specifications get a str filter and return a new one.
        That is why the initial_query is an empty str.
        """
        return ''

    def get(self, *specifications: Specification, lazy: bool = False):
        key_name = self.apply_specifications(specifications)

        if lazy:
            return self.lazy_command_cls(self.session.get, key_name)
        else:
            return self.model.from_json(self.session.get(key_name))

    def filter(self, *specifications: Specification, lazy: bool = False):
        key_name = self.apply_specifications(specifications)

        if lazy:
            return self.lazy_command_cls(self.session.keys, key_name)

        keys = self.session.keys(key_name)
        return [self.model.from_json(value) for value in self.session.mget(*keys)]

    def save(self, obj: RedisModel):
        self.session.set(str(obj.id), obj.json())

    def update_many(self, specifications: Iterable[Specification], updated_fields):
        raise NotImplementedError()     # TODO: implement

    def delete(self, obj: RedisModel):
        self.session.delete(str(obj.id))

    def update(self, obj: RedisModel):
        self.save(obj)

    def is_modified(self, obj: RedisModel):
        return self.get(FilterSpecification(obj.id), lazy=False) == obj

    def refresh(self, obj: RedisModel):
        fresh_obj = self.get(FilterSpecification(obj.id), lazy=False)

        for key, value in fresh_obj.dict().items():
            setattr(obj, key, value)



import json
from collections.abc import Iterable

from kafka import KafkaConsumer, KafkaProducer
from kafka.errors import KafkaError

from assimilator.core.events import Event, UnknownEvent
from assimilator.core.events import EventParsingError, EventProducingError
from assimilator.core.events.events_bus import EventConsumer, EventProducer


class KafkaEventConsumer(EventConsumer):
    def __init__(self, topics: Iterable[str], consumer: KafkaConsumer):
        self.consumer = consumer
        self.topics = list(topics)

    def close(self):
        self.consumer.close()

    def start(self):
        """ Connected by default """

    def consume(self) -> Iterable[UnknownEvent]:
        self.consumer.subscribe(self.topics)

        for message in self.consumer:
            try:
                yield UnknownEvent.from_json(message.value)
            except json.JSONDecoder as exc:
                raise EventParsingError(exc)


class KafkaEventProducer(EventProducer):
    def __init__(self, topic: str, producer: KafkaProducer, sync_produce: bool = False, timeout: int = None):
        self.topic = topic
        self.producer = producer
        self.sync_produce = sync_produce
        self.timeout = timeout

    def produce(self, event: Event):
        message = self.producer.send(self.topic, key=event.id, value=event.json())

        if self.sync_produce:
            try:
                message.get(timeout=self.timeout)
            except KafkaError as exc:
                raise EventProducingError(exc)

    def start(self):
        """ Already started """

    def close(self):
        self.producer.close()



from contextlib import contextmanager

import assimilator.core
import assimilator.internal


@contextmanager
def optional_dependencies(error: str = "ignore"):
    assert error in {"raise", "warn", "ignore"}
    try:
        yield None
    except ImportError as e:
        if error == "raise":
            raise e
        if error == "warn":
            msg = f'Missing optional dependency "{e.name}". Use pip or conda to install.'
            print(f'Warning: {msg}')


with optional_dependencies():
    import assimilator.alchemy

with optional_dependencies():
    import assimilator.kafka


with optional_dependencies():
    import assimilator.redis



from assimilator.redis.database.repository import *
from assimilator.redis.database.models import *
from assimilator.redis.database.specifications import *
from assimilator.redis.database.unit_of_work import *



from assimilator.core.database.unit_of_work import UnitOfWork


class RedisUnitOfWork(UnitOfWork):
    _saved_session = None

    def begin(self):
        self._saved_session = self.repository.session
        self.repository.session = self.repository.session.pipeline()

    def rollback(self):
        self.repository.session.discard()

    def commit(self):
        self.repository.session.execute()

    def close(self):
        self.repository.session.reset()
        self.repository.session = self._saved_session
        self._saved_session = None



from abc import ABC, abstractmethod


class StartCloseContextMixin(ABC):
    @abstractmethod
    def close(self):
        raise NotImplementedError("close() is not implemented")

    @abstractmethod
    def start(self):
        raise NotImplementedError("start() is not implemented")

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

        if exc_type is not None:
            raise exc_val



from sqlalchemy import Column, BigInteger, Text, DateTime

from assimilator.core.events import Event
from assimilator.core.database.unit_of_work import UnitOfWork
from assimilator.core.events import OutboxRelay
from assimilator.core.events.events_bus import EventBus


def create_outbox_event_model(Base):
    class OutboxEvent(Base):
        id = Column(BigInteger())
        event_data = Column(Text())
        event_date = Column(DateTime(timezone=True))

        def __init__(self, event: Event, *args, **kwargs):
            super(OutboxEvent, self).__init__(
                event_data=event.json(),
                event_date=event.event_date,
                *args,
                **kwargs,
            )

    return OutboxEvent


class AlchemyOutboxRelay(OutboxRelay):
    def __init__(self, outbox_event_model, uow: UnitOfWork, event_bus: EventBus):
        self.outbox_event_model = outbox_event_model
        super(AlchemyOutboxRelay, self).__init__(uow=uow, event_bus=event_bus)

    def start(self):
        while True:
            with self.uow:
                events = self.uow.repository.filter()

                for event in events:
                    self.event_bus.produce(event)

                self.acknowledge(events)
                self.uow.commit()

            self.delay_function()

    def delay_function(self):
        raise NotImplementedError("delay function is not implemented")

    def acknowledge(self, events):
        for event in events:
            self.uow.repository.delete(event)
